package grpc

import (
	"reflect"
	"testing"

	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"git.perx.ru/perxis/perxis-go/proto/common"
	"github.com/hashicorp/go-multierror"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func Test_getBadRequest(t *testing.T) {
	tests := []struct {
		name string
		err  error
		want *common.Error_BadRequest
	}{
		{"Without field error", errors.New("some"), nil},
		{"Field error", errors.WithField(errors.New("some"), "fld1"),
			&common.Error_BadRequest{Errors: []*common.Error_BadRequest_FieldViolation{{Description: "some", Field: "fld1"}}}},
		{"Deep field error", errors.Wrap(errors.WithField(errors.New("some"), "fld1"), "extra"),
			&common.Error_BadRequest{Errors: []*common.Error_BadRequest_FieldViolation{{Description: "some", Field: "fld1"}}}},
		{"Nested field errors", errors.WithField(errors.WithField(errors.New("some"), "fld1"), "fld2"),
			&common.Error_BadRequest{Errors: []*common.Error_BadRequest_FieldViolation{{Description: "some", Field: "fld2.fld1"}}}},
		{"Deep nested field errors", errors.WithField(errors.Wrap(errors.WithField(errors.New("some"), "fld1"), "extra"), "fld2"),
			&common.Error_BadRequest{Errors: []*common.Error_BadRequest_FieldViolation{{Description: "some", Field: "fld2.fld1"}}}},
		{"Multi Field error", multierror.Append(errors.WithField(errors.New("some"), "fld1"), errors.WithField(errors.New("some"), "fld2")),
			&common.Error_BadRequest{Errors: []*common.Error_BadRequest_FieldViolation{{Description: "some", Field: "fld1"}, {Description: "some", Field: "fld2"}}}},
		{"Nested Multi Field error", errors.WithField(multierror.Append(errors.WithField(errors.New("some"), "fld1"), errors.WithField(errors.New("some"), "fld2")), "fld3"),
			&common.Error_BadRequest{Errors: []*common.Error_BadRequest_FieldViolation{{Description: "some", Field: "fld3.fld1"}, {Description: "some", Field: "fld3.fld2"}}}},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if got := getBadRequest(tt.err); !reflect.DeepEqual(got, tt.want) {
				t.Errorf("getBadRequest() = %v, want %v", got, tt.want)
			}
		})
	}
}

func Test_getHelp(t *testing.T) {
	tests := []struct {
		name string
		err  error
		want *common.Error_Help
	}{
		{"Without help", errors.New("some"), nil},
		{"With help",
			errors.WithHelp(errors.New("some"), &errors.Help{Links: []errors.HelpLink{{Description: "Help1", URL: "Url1"}}}),
			&common.Error_Help{Links: []*common.Error_Help_Link{{Description: "Help1", Url: "Url1"}}}},
		{"With nested help",
			errors.WithHelp(errors.WithHelp(errors.New("some"), &errors.Help{Links: []errors.HelpLink{{Description: "Help2", URL: "Url2"}}}), &errors.Help{Links: []errors.HelpLink{{Description: "Help1", URL: "Url1"}}}),
			&common.Error_Help{Links: []*common.Error_Help_Link{{Description: "Help1", Url: "Url1"}, {Description: "Help2", Url: "Url2"}}}},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			if got := getHelp(tt.err); !reflect.DeepEqual(got, tt.want) {
				t.Errorf("getHelp() = %v, want %v", got, tt.want)
			}
		})
	}
}

func TestProtoFromError(t *testing.T) {
	err := errors.New("some error")
	err = errors.SetID(err, "0")
	err = errors.WithDetail(err, "Ошибка")
	err = errors.WithDomain(err, "domain-1")
	err = errors.WithHelp(err, &errors.Help{Links: []errors.HelpLink{{Description: "Docs", URL: "https://docs.com"}}})
	err = errors.WithErrors(err,
		errors.SetID(errors.WithDetail(errors.New("err-1"), "Ошибка-1"), "1"),
		errors.SetID(errors.WithDetail(errors.New("err-2"), "Ошибка-2"), "2"),
		errors.SetID(errors.WithDetail(errors.New("err-3"), "Ошибка-3"), "3"),
	)

	assert.Equal(t,
		&common.Error{
			ErrorCode: 0,
			ErrorId:   "0",
			Reason:    "",
			Domain:    "domain-1",
			Message:   "some error",
			Details:   "Ошибка",
			Help: &common.Error_Help{
				Links: []*common.Error_Help_Link{{
					Description: "Docs",
					Url:         "https://docs.com",
				}},
			},
			Errors: []*common.Error{
				{ErrorId: "1", Message: "err-1", Details: "Ошибка-1"},
				{ErrorId: "2", Message: "err-2", Details: "Ошибка-2"},
				{ErrorId: "3", Message: "err-3", Details: "Ошибка-3"},
			},
		},
		ProtoFromError(err),
	)
}

func TestErrorFromProto(t *testing.T) {
	pberr := &common.Error{
		ErrorCode: 0,
		ErrorId:   "0",
		Reason:    "",
		Domain:    "domain-1",
		Message:   "some error",
		Details:   "Ошибка",
		Help: &common.Error_Help{
			Links: []*common.Error_Help_Link{{
				Description: "Docs",
				Url:         "https://docs.com",
			}},
		},
		Errors: []*common.Error{
			{ErrorId: "1", Message: "err-1", Details: "Ошибка-1"},
			{ErrorId: "2", Message: "err-2", Details: "Ошибка-2"},
			{ErrorId: "3", Message: "err-3", Details: "Ошибка-3"},
		},
	}

	err := ErrorFromProto(nil, pberr)
	require.NotNil(t, err)
	assert.Equal(t, "0", errors.GetID(err))
	assert.Equal(t, "some error", err.Error())
	assert.Equal(t, "Ошибка", errors.GetDetail(err))
	assert.Equal(t, "domain-1", errors.GetDomain(err))
	assert.Equal(t, &errors.Help{Links: []errors.HelpLink{{Description: "Docs", URL: "https://docs.com"}}}, errors.GetHelp(err))

	errs := errors.GetErrors(err)
	require.Len(t, errs, 3)

	assert.Equal(t, "1", errors.GetID(errs[0]))
	assert.Equal(t, "err-1", errs[0].Error())
	assert.Equal(t, "Ошибка-1", errors.GetDetail(errs[0]))

	assert.Equal(t, "2", errors.GetID(errs[1]))
	assert.Equal(t, "err-2", errs[1].Error())
	assert.Equal(t, "Ошибка-2", errors.GetDetail(errs[1]))

	assert.Equal(t, "3", errors.GetID(errs[2]))
	assert.Equal(t, "err-3", errs[2].Error())
	assert.Equal(t, "Ошибка-3", errors.GetDetail(errs[2]))
}
